@@ -10,8 +10,8 @@ from mock import patch, Mock, MagicMock
 from unittest import TestCase
 from voluptuous import Schema, Optional, Invalid as InvalidSchema
 import pystmark
-from pystmark import (Sender, BatchSender, Message,
-                      UnauthorizedError, UnprocessableEntityError,
+from pystmark import (Sender, BatchSender, TemplateSender, BatchTemplateSender,
+                      Message, UnauthorizedError, UnprocessableEntityError,
                       InternalServerError, MessageError,
                       MAX_RECIPIENTS_PER_MESSAGE, MAX_BATCH_MESSAGES,
                       POSTMARK_API_TEST_KEY, POSTMARK_API_URL,
@@ -180,7 +180,7 @@ class SenderTestBase(TestCase):
         return self.sender.send(message=self.message)
 
 
-class SenderTest(SenderTestBase):
+class SenderTestMixin(object):
 
     @patch.object(requests.Session, 'request', autospec=True)
     def test_send_message(self, mock_request):
@@ -241,6 +241,28 @@ class SenderTest(SenderTestBase):
         self.assertEqual(json.loads(request_data)['TextBody'], test_text)
 
 
+class SenderTest(SenderTestBase, SenderTestMixin):
+    pass
+
+
+class TemplateSenderTestBase(SenderTestBase):
+    @property
+    def sender(self):
+        return TemplateSender(test=True)
+
+    def send(self):
+        return self.sender.send(message=self.message)
+
+
+class TemplateSenderTest(TemplateSenderTestBase, SenderTestMixin):
+
+    @patch.object(requests.Session, 'request')
+    def test_send_with_template(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        r = pystmark.send_with_template(self.message, test=True)
+        self.assertValidJSONResponse(r, self.schema)
+
+
 class ResponseTest(SenderTestBase):
 
     @patch.object(requests.Session, 'request', autospec=True)
@@ -338,15 +360,22 @@ class MessageTest(SenderTestBase):
 
     def test_load_message_native(self):
         msg = dict(to='me', text='hi', html='<b>hi</b>', reply_to='you',
-                   cc='dog,cat', subject='dogs', headers=[dict(Name='Food',
-                                                               Value='7')])
+                   cc='dog,cat', bcc='foo,bar', subject='dogs',
+                   track_opens=True, headers=[dict(Name='Food', Value='7')],
+                   attachments=[], sender='admin', tag='tag',
+                   template_id='template_id', template_alias='template_alias',
+                   template_model='template_model',
+                   message_stream='message_stream')
+        self.assertEqual(sorted(msg), sorted(Message._fields))
         self.assertNotRaises(TypeError, Message.load_message, msg)
         self.assertNotRaises(MessageError, Message.load_message, msg,
                              verify=True)
+
         msg = dict(to='me', text='hi')
         self.assertNotRaises(TypeError, Message.load_message, msg)
         self.assertNotRaises(MessageError, Message.load_message, msg,
                              verify=True)
+
         pystmsg = Message.load_message(msg)
         self.assertEqual(pystmsg.data(), dict(To='me', TextBody='hi'))
 
@@ -575,7 +604,6 @@ class MessageErrorTest(SenderTestBase):
         self.assertRaises(MessageError, m._detect_content_type, 'xxx.bin')
         # Unknown extension returns default content type
         ext = 'xcacaswcawc'
-        m._allowed_extensions.append(ext)
         self.assertEqual(m._detect_content_type('xxx.' + ext),
                          m._default_content_type)
         # Known extension returns correct mimetype
@@ -735,6 +763,80 @@ class BatchSenderTest(BatchSenderTestBase):
         self.assertValidJSONResponse(r, self.schema)
 
 
+class BatchTemplateSenderTestBase(SenderTestBase):
+
+    response = [SenderTestBase.response] * 10
+
+    schema = [SenderTestBase.schema]
+
+    _message_count = 20
+
+    def setUp(self):
+        super(BatchTemplateSenderTestBase, self).setUp()
+        self._messages = None
+
+    @property
+    def message_count(self):
+        return self._message_count
+
+    @message_count.setter
+    def message_count(self, v):
+        if v != self._message_count:
+            self._messages = None
+        self._message_count = v
+
+    @property
+    def messages(self):
+        if self._messages is not None:
+            return self._messages
+        msgs = [{}] * self.message_count
+        [msg.update(self.message) for msg in msgs]
+        for msg in msgs:
+            msg['To'] = '{0}@example.com'.format(_make_random_string(10))
+        print(msgs)
+        return msgs
+
+    @property
+    def sender(self):
+        return BatchTemplateSender(api_key=POSTMARK_API_TEST_KEY, test=True)
+
+    def send(self):
+        return self.sender.send(messages=self.messages)
+
+
+class BatchTemplateSenderTest(BatchTemplateSenderTestBase):
+
+    @patch.object(requests.Session, 'request')
+    def test_send_batch_templates(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        r = self.send()
+        self.assertValidJSONResponse(r, self.schema)
+
+    @patch.object(requests.Session, 'request')
+    def test_send_batch_templates_no_messages(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        msg = 'No messages to send'
+        self.assertRaisesMessage(MessageError, msg, self.sender.send,
+                                 messages=None)
+        self.assertRaisesMessage(MessageError, msg, self.sender.send,
+                                 messages=[])
+
+    @patch.object(requests.Session, 'request')
+    def test_send_batch_templates_too_many_messages(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        self.message_count = MAX_BATCH_MESSAGES + 1
+        msg = 'Maximum {0} messages allowed in batch'
+        msg = msg.format(MAX_BATCH_MESSAGES)
+        self.assertRaisesMessage(MessageError, msg, self.sender.send,
+                                 messages=self.messages)
+
+    @patch.object(requests.Session, 'request')
+    def test_simple_api(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        r = pystmark.send_batch_with_templates(self.messages, test=True)
+        self.assertValidJSONResponse(r, self.schema)
+
+
 class BouncesTest(TestCase):
 
     response = {
@@ -894,6 +996,91 @@ class BounceTest(TestCase):
         self.assertIs(r.bounce, None)
 
 
+class OutboundMessageDetailsTest(TestCase):
+
+    message_id = "00000000-0000-0000-0000-000000000000"
+    response = {
+          "TextBody": "Thank you for your order...",
+          "HtmlBody": "<p>Thank you for your order...</p>",
+          "Body": "SMTP dump data",
+          "Tag": "product-orders",
+          "MessageID": "07311c54-0687-4ab9-b034-b54b5bad88ba",
+          "To": [
+            {
+              "Email": "john.doe@yahoo.com",
+              "Name": None
+            }
+          ],
+          "Cc": [],
+          "Bcc": [],
+          "Recipients": [
+            "john.doe@yahoo.com"
+          ],
+          "ReceivedAt": "2014-02-14T11:12:54.8054242-05:00",
+          "From": "\"Joe\" <joe@domain.com>",
+          "Subject": "Parts Order #5454",
+          "Attachments": [
+              "myimage.png",
+              "mypaper.doc"
+          ],
+          "Metadata": {
+            "color": "blue",
+            "client-id": "12345"
+          },
+          "Status": "Sent",
+          "TrackOpens": True,
+          "TrackLinks": "HtmlOnly",
+          "MessageEvents": [
+            {
+              "Recipient": "john.doe@yahoo.com",
+              "Type": "Delivered",
+              "ReceivedAt": "2014-02-14T11:13:10.8054242-05:00",
+              "Details": {
+                "DeliveryMessage":
+                    "smtp;250 2.0.0 OK l10si21599969igu.63 - gsmtp",
+                "DestinationServer":
+                    "yahoo-smtp-in.l.yahoo.com (433.899.888.26)",
+                "DestinationIP": "173.194.74.256"
+              }
+            },
+            {
+              "Recipient": "john.doe@yahoo.com",
+              "Type": "Opened",
+              "ReceivedAt": "2014-02-14T11:20:10.8054242-05:00",
+              "Details": {
+                "Summary": "Email opened with Mozilla/5.0"
+              }
+            }
+          ]
+    }
+    schema = {
+        "TextBody": unicode,
+        "HtmlBody": unicode,
+        "Body": unicode,
+        "Tag": unicode,
+        "MessageID": unicode,
+        "To": list,
+        "Cc": list,
+        "Bcc": list,
+        "Recipients": list,
+        "ReceivedAt": unicode,
+        "From": unicode,
+        "Subject": unicode,
+        "Attachments": list,
+        "Metadata": dict,
+        "Status": unicode,
+        "TrackOpens": bool,
+        "TrackLinks": unicode,
+        "MessageEvents": list
+    }
+
+    @patch.object(requests.Session, 'request')
+    def test_simple_api(self, mock_request):
+        mock_request.return_value = self.mock_response(self.json_response)
+        r = pystmark.get_outbound_message_details(self.message_id, test=True)
+        self.assertValidJSONResponse(r, self.schema)
+
+
 class BounceDumpTest(TestCase):
 
     bounce_id = 777
